! Example program: TTT -- A very simple Tic Tac Toe game

! This intention with this program is to provide a first
! overview over the T3X language. Everything fllowing an
! exclamation point (like these lines) is ignored by the
! compiler.

! This module is called 'ttt' and depends on the T3X core class
! (which virtually all T3X programs do) and on the TTY control
! class, which allows to move the cursor, query the keyboard, etc.
module ttt(t3x, ttyctl);

! Declare instances of the used classes so that messages can be
! sent to them.
object	t[t3x], tty[ttyctl];

! This is a 9-element integer vector holding the playing field.
var	Field[9];

! This variable will be initialized with the move database. See init().
var	Moves;

! This is a procedure declaration with a block statement (DO...END)
! as its body.
! It is used to set up variables, initialize objects, etc.
init() do
	! Rules for winning TTT
	! 0 = empty  1 = human  2 = computer
	! lefthand side = pattern   righthand side = move
	! priority is descending 
	! The value, Moves is initialized with, is a so-called
	! Table. In T3X, a Table is a collection of things like
	! numbers, strings, letters, etc, enclosed by square
	! brackets. A table may in turn contain tables, like
	! the below one.
	Moves := [
		[ [0,2,2], [2,2,2] ],	! Winning moves
		[ [2,0,2], [2,2,2] ],
		[ [2,2,0], [2,2,2] ],
		[ [0,1,1], [2,1,1] ],	! Blocking moves
		[ [1,0,1], [1,2,1] ],
		[ [1,1,0], [1,1,2] ],
		[ [0,1,0], [2,1,0] ],	! Preventive moves
		[ [0,0,1], [0,2,1] ],
		[ [1,0,0], [1,2,0] ],
		[ [0,1,2], [2,1,2] ],	! Filling moves
		[ [0,2,1], [2,2,1] ],
		[ [1,0,2], [1,2,2] ],
		[ [1,2,0], [1,2,2] ],
		[ [2,0,1], [2,2,1] ],
		[ [2,1,0], [2,1,2] ],
		[ [0,0,2], [2,0,2] ],	! Building moves
		[ [0,2,0], [2,2,0] ],
		[ [2,0,0], [2,0,2] ],
		[ [0,0,0], [0,2,0] ],	! First move (unused)
		0
	];
	! Initialize the TTY object. There are no constructors in T3X,
	! but everything must be done explicitly.
	! The TTYCTL class is a dynamically loadable class, so the
	! init() method loads it into RAM (if not yet present).
	! The dot is used to send a message to an object.
	! TTY.INIT() means 'send INIT() to TTY'.
	tty.init();
	tty.mode(1);	! Set up the TTY for processing single keystrokes
	tty.clear();	! Clear the terminal screen.
	t.memfill(Field, 0, 9*t.bpw());	! Clear the playing field.
end

! Shut down everything and exit.
fini() do
	tty.move(0, 12);	! Move the cursor to below the field.
	tty.mode(0);		! Restore original TTY state.
	tty.fini();		! Unload the TTY routines.
	halt 0;			! Halt program execution.
end

! Draw the frame of the playing field.
! TTY.move(x,y) moves the cursor.
! TTY.writes(s) writes a packed string.
! When a string is not prefixed with the keyword PACKED,
! it allocates one word per character. Packed strings
! allocate one byte per character. The TTYCTL methods
! (like most routines) expect packed strings.
drawframe() do
	tty.move(8, 4); tty.writes(packed"   |   |   ");
	tty.move(8, 5); tty.writes(packed"---+---+---");
	tty.move(8, 6); tty.writes(packed"   |   |   ");
	tty.move(8, 7); tty.writes(packed"---+---+---");
	tty.move(8, 8); tty.writes(packed"   |   |   ");
end

! Redraw the playing field and all figures already placed.
drawfield() do
	var i, j;	! the variables are local to the DO ... END block

	! Perform the body (DO...END) for j=0, j=1, j=2 .
	for (j=0, 3) do
		for (i=0, 3) do
			! Move to field with coordinates x=i, y=j.
			tty.move(9+4*i, 4+2*j);
			tty.writec(
				! If the field contains 1, write 'X'
				Field[j*3+i]=1->'X':
				! If the field contains 2, write 'O'
				Field[j*3+i]=2-> 'O':
				! Else write a space character
				'\s');
		end
	end
end


! Match a row given by the cells fp0[0], fp1[0], fp2[0] against
! the patterns in the Moves table.
! If a matching entry is found, replace the cells with the
! 'move' pattern associated with found entry.
match(fp0, fp1, fp2, rule) do var p, q;
	p := rule[0];	! Pattern
	q := rule[1];	! Move
	! If all three cells are equal, a match is found.
	if (fp0[0] = p[0] /\ fp1[0] = p[1] /\ fp2[0] = p[2]) do
		! In this case, replace it with the resulting move...
		fp0[0] := q[0];
		fp1[0] := q[1];
		fp2[0] := q[2];
		! ... and return 1 for success
		return 1;
	end
	! No match: return 0.
	! Any procedure, which does not explicitly return a
	! value, returns zero.
end

! Do a move.
! N is the field selected by the (human) player.
move(n) do var i;
	Field[n] := 1;		! Place human player's figure.
	if (Field[4] = 0) do	! Is the center unoccupied?
		Field[4] := 2;		! If so, occupy it
		drawfield();		!  and accept the next move
		return;
	end
	! Find row in the playing field which matches a pattern
	! in the pattern/move table (Moves).
	! Begin with the zero'th slot of Moves (Moves[0]) and
	! continue until Moves[i]=0m, which indicates that all
	! patterns have been tried.
	! When a match was successful, leave the loop. match()
	! performs the move in case of success.
	! Notice: under normal conditions, the end of the table
	! cannot be reached.
	i := 0; while (Moves[i]) do
		! Diagonal matches
		if (match(@Field[0], @Field[4], @Field[8], Moves[i])) leave;
		if (match(@Field[2], @Field[4], @Field[6], Moves[i])) leave;
		! Horizontal matches
		if (match(@Field[0], @Field[1], @Field[2], Moves[i])) leave;
		if (match(@Field[3], @Field[4], @Field[5], Moves[i])) leave;
		if (match(@Field[6], @Field[7], @Field[8], Moves[i])) leave;
		! Vertical matches
		if (match(@Field[0], @Field[3], @Field[6], Moves[i])) leave;
		if (match(@Field[1], @Field[4], @Field[7], Moves[i])) leave;
		if (match(@Field[2], @Field[5], @Field[8], Moves[i])) leave;
		i := i+1;
	end
	drawfield();
	if (i < 3) do		! Made a winning move?
		tty.move(8, 10);
		tty.writes(packed"Oops!");
		fini();
	end
end

! Let the player select a field. Accept only free cells.
select() do var i, j, k;
	! Find the first free cell (Field[i]=0).
	for (i=0, 9) if (\Field[i]) leave;
	j := i / 3;	! j = row
	i := i mod 3;	! i = column
	! Start an endless loop which is left (using LEAVE)
	! when the player selected a free cell.
	while (1) do
		tty.move(9+4*i, 4+2*j);	! Move cursor to current cell.
		k := tty.readc();	! Read single key from TTY
		! If one of the cursor keys (up,down,left,right) was
		! pressed, move the cursor accordingly
		ie (k = 'l' \/ k = ttyctl.K_RIGHT) do
			i := i+1;
			if (i > 2) i := 0;
		end
		else ie (k = 'h' \/ k = ttyctl.K_LEFT) do
			i := i-1;
			if (i < 0) i := 2;
		end
		else ie (k = 'j' \/ k = ttyctl.K_DOWN) do
			j := j+1;
			if (j > 2) j := 0;
		end
		else ie (k = 'k' \/ k = ttyctl.K_UP) do
			j := j-1;
			if (j < 0) j := 2;
		end
		! ENTER pressed: select the current cell.
		else ie (k = '\r') do
			ie (Field[j*3+i])
				tty.writec('\a'); ! already occupied - BEEP
			else
				leave;		  ! free - leave loop
		end
		! Pressing ESCape (or 'q') terminates the program instantly
		else if (k = ttyctl.K_ESC \/ k = 'q') do
			fini();
		end
	end
	return j*3+i;	! return the selected box number
end

! This is the main program. Execution begins here.
do var i;
	init();				! Initialize data and TTY.
	drawframe();			! Draw field.
	for (i=0, 5) move(select());	! Play (max. 5 moves).
	fini();				! Shut down everything and exit.
end
